import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firestore_repository_helper/src/firestore_object_extensions.dart';
import 'package:resource_result/resource_result.dart';
import 'package:future_helper/future_helper.dart';
import 'package:serializer_interface/serializer_interface.dart';

import 'firestore_object.dart';
import 'no_data_failure.dart';
import 'snapshot_serialization_exception.dart';

abstract class FirestoreRepository<T extends FirestoreObject<T>> {
  String get collectionPath;

  JsonSerializer<T> get serializer;

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if documents do not exist
  Future<Resource<List<T>>> readAll();

  Future<Resource<List<T>>> query(
      CollectionReference<T> Function(CollectionReference<T> collection)
          applyQuery);

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if document does not exist
  Future<Resource<T>> readDocument(String id);

  /// throws [Exception] if there was an error.
  Future updateDocument(T data);

  /// throws [Exception] if there was an error.
  Future deleteDocument(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if document does not exist
  Stream<Resource<T>> listenToDocument(String id);
}

abstract class FirestoreRepositoryImpl<T extends FirestoreObject<T>>
    implements FirestoreRepository<T> {
  CollectionReference<T> _getCollectionRef() {
    // for converter see: https://firebase.flutter.dev/docs/firestore/usage/#typing-collectionreference-and-documentreference
    return FirebaseFirestore.instance
        .collection(collectionPath)
        .withConverter<T>(
            fromFirestore: fromSnapshotConverter,
            toFirestore: toSnapShotConverter);
  }

  T fromSnapshotConverter(DocumentSnapshot<Map<String, dynamic>> snapshot,
      SnapshotOptions? options) {
    try {
      return serializer.fromJson(snapshot.data()!).withId(snapshot.id);
    } catch (e) {
      throw SerializationException.fromJson(snapshot.data());
    }
  }

  Map<String, dynamic> toSnapShotConverter(
      T firestoreObj, SetOptions? options) {
    try {
      return serializer.toJson(firestoreObj);
    } catch (e) {
      throw SerializationException.toJson(firestoreObj);
    }
  }

  @override
  Future deleteDocument(String id) => _getCollectionRef().doc(id).delete();

  Resource<T> fromSnapshot(DocumentSnapshot<T> snapshot) {
    if (!snapshot.exists || snapshot.data() == null) {
      return NoDataFailure(
          Error('No document with id "${snapshot.id}" with data found.'));
    }
    return Success(snapshot.data()!);
  }

  @override
  Future<Resource<List<T>>> readAll() async {
    final snapshots = await _getCollectionRef().get();
    final objects = snapshots.docs.map((snapshot) => snapshot.data()).toList();
    return Success(objects);
  }

  @override
  Future<Resource<List<T>>> query(
      Query<T> Function(CollectionReference<T> collection) applyQuery) async {
    final snapshots = await applyQuery(_getCollectionRef()).get();
    final objects = snapshots.docs.map((snapshot) => snapshot.data()).toList();
    return Success(objects);
  }

  @override
  Future<Resource<T>> readDocument(String id) async {
    return _getCollectionRef().doc(id).get().map(fromSnapshot);
  }

  @override
  Future updateDocument(T data) =>
      _getCollectionRef().doc(data.id).update(serializer.toJson(data));

  @override
  Stream<Resource<T>> listenToDocument(String id) {
    return _getCollectionRef().doc(id).snapshots().map(fromSnapshot);
  }
}
